#include "ecdsa.h"
#include "utilities.h"

#include <openssl/sha.h>
#include <openssl/hmac.h>
#include <string.h>

static const EC_METHOD* CurveTypeToMethod(ECCurveType curveType)
{
	if (curveType == PrimeShortWeierstrass)
		return EC_GFp_simple_method();

	if (curveType == PrimeMontgomery)
		return EC_GFp_mont_method();

#if HAVE_OPENSSL_EC2M
	if (API_EXISTS(EC_GF2m_simple_method) && (curveType == Characteristic2))
		return EC_GF2m_simple_method();
#endif

	return NULL; //Edwards and others
}

EC_KEY* CryptoNative_EcKeyCreateByExplicitParameters(
	ECCurveType curveType,
	uint8_t* qx, int32_t qxLength,
	uint8_t* qy, int32_t qyLength,
	uint8_t* d, int32_t dLength,
	uint8_t* p, int32_t pLength,
	uint8_t* a, int32_t aLength,
	uint8_t* b, int32_t bLength,
	uint8_t* gx, int32_t gxLength,
	uint8_t* gy, int32_t gyLength,
	uint8_t* order, int32_t orderLength,
	uint8_t* cofactor, int32_t cofactorLength,
	uint8_t* seed, int32_t seedLength)
{
	if (!p || !a || !b || !gx || !gy || !order || !cofactor)
	{
		// qx, qy, d and seed are optional
		assert(false);
		return 0;
	}

	EC_KEY* key = NULL;
	EC_POINT* G = NULL;

	BIGNUM* qxBn = NULL;
	BIGNUM* qyBn = NULL;
	BIGNUM* dBn = NULL;
	BIGNUM* pBn = NULL; // p = either the char2 polynomial or the prime
	BIGNUM* aBn = NULL;
	BIGNUM* bBn = NULL;
	BIGNUM* gxBn = NULL;
	BIGNUM* gyBn = NULL;
	BIGNUM* orderBn = NULL;
	BIGNUM* cofactorBn = NULL;

	// Create the group. Explicitly specify the curve type because using EC_GROUP_new_curve_GFp
	// will default to montgomery curve
	const EC_METHOD* curveMethod = CurveTypeToMethod(curveType);
	if (!curveMethod) return NULL;

	EC_GROUP* group = EC_GROUP_new(curveMethod);
	if (!group) return NULL;

	pBn = BN_bin2bn(p, pLength, NULL);
	// At this point we should use 'goto error' since we allocated memory
	aBn = BN_bin2bn(a, aLength, NULL);
	bBn = BN_bin2bn(b, bLength, NULL);

#if HAVE_OPENSSL_EC2M
	if (API_EXISTS(EC_GROUP_set_curve_GF2m) && (curveType == Characteristic2))
	{
		if (!EC_GROUP_set_curve_GF2m(group, pBn, aBn, bBn, NULL))
			goto error;
	}
	else
#endif
	{
		if (!EC_GROUP_set_curve_GFp(group, pBn, aBn, bBn, NULL))
			goto error;
	}

	// Set generator, order and cofactor
	G = EC_POINT_new(group);
	gxBn = BN_bin2bn(gx, gxLength, NULL);
	gyBn = BN_bin2bn(gy, gyLength, NULL);

#if HAVE_OPENSSL_EC2M
	if (API_EXISTS(EC_POINT_set_affine_coordinates_GF2m) && (curveType == Characteristic2))
	{
		EC_POINT_set_affine_coordinates_GF2m(group, G, gxBn, gyBn, NULL);
	}
	else
#endif
	{
		EC_POINT_set_affine_coordinates_GFp(group, G, gxBn, gyBn, NULL);
	}

	orderBn = BN_bin2bn(order, orderLength, NULL);
	cofactorBn = BN_bin2bn(cofactor, cofactorLength, NULL);
	EC_GROUP_set_generator(group, G, orderBn, cofactorBn);

	// Set seed (optional)
	if (seed && seedLength > 0)
	{
		if (!EC_GROUP_set_seed(group, seed, (size_t)seedLength))
			goto error;
	}

	// Validate group
	if (!EC_GROUP_check(group, NULL))
		goto error;

	// Create key
	key = EC_KEY_new();
	if (!key)
		goto error;

	if (!EC_KEY_set_group(key, group))
		goto error;

	// Set the public and private key values
	if (qx && qy)
	{
		qxBn = BN_bin2bn(qx, qxLength, NULL);
		qyBn = BN_bin2bn(qy, qyLength, NULL);
		if (!qxBn || !qyBn)
			goto error;

		if (!EC_KEY_set_public_key_affine_coordinates(key, qxBn, qyBn))
			goto error;

		// Set private key (optional)
		if (d && dLength)
		{
			dBn = BN_bin2bn(d, dLength, NULL);
			if (!dBn)
				goto error;

			if (!EC_KEY_set_private_key(key, dBn))
				goto error;
		}

		// Validate key
		if (!EC_KEY_check_key(key))
			goto error;
	}

	// Success
	return key;

error:
	if (qxBn) BN_free(qxBn);
	if (qyBn) BN_free(qyBn);
	if (dBn) BN_free(dBn);
	if (pBn) BN_free(pBn);
	if (aBn) BN_free(aBn);
	if (bBn) BN_free(bBn);
	if (gxBn) BN_free(gxBn);
	if (gyBn) BN_free(gyBn);
	if (orderBn) BN_free(orderBn);
	if (cofactorBn) BN_free(cofactorBn);
	if (G) EC_POINT_free(G);
	if (group) EC_GROUP_free(group);
	if (key) EC_KEY_free(key);
	return NULL;
}

int32_t
CryptoNative_EcDsaSignEx(
	const uint8_t* dgst, int32_t dgstlen,
	uint8_t* sig, int32_t* siglen,
	uint8_t* kinv, int32_t kinvLength,
	uint8_t* rp, int32_t rpLength,
	EC_KEY* key)
{
	if (!siglen)
	{
		return 0;
	}
	BIGNUM* kinvBn = NULL;
	BIGNUM* rpBn = NULL;

	kinvBn = BN_bin2bn(kinv, kinvLength, NULL);
	rpBn = BN_bin2bn(rp, rpLength, NULL);

	unsigned int unsignedSigLength = Int32ToUint32(*siglen);
	int ret = ECDSA_sign_ex(0, dgst, dgstlen, sig, &unsignedSigLength, kinvBn, rpBn, key);
	*siglen = Uint32ToInt32(unsignedSigLength);

	if (kinvBn) BN_free(kinvBn);
	if (rpBn) BN_free(rpBn);
	
	return ret;
}

int32_t CryptoNative_EcDsaSize(const EC_KEY* key)
{
	return ECDSA_size(key);
}

int32_t
CryptoNative_EcPspParameter(const EC_KEY* key, const uint8_t* sha256, int32_t sha256len, BIGNUM** kinv, int32_t* kinvlen, BIGNUM** rp, int32_t* rplen)
{
	if (!key || ! sha256 ||
		!kinv || !kinvlen ||
		!rp || !rplen)
	{
		assert(false);

		if (kinv) *kinv = NULL; if (kinvlen) *kinvlen = 0;
		if (rp) *rp = NULL; if (rplen) *rplen = 0;
		return 0;
	}

	int32_t rc = 0;
	
	BIGNUM* m = NULL;
	BIGNUM* orderBn = NULL;
	BIGNUM* exp = NULL;
	BIGNUM* sig_r = NULL;
	BIGNUM* minv = NULL;

	BN_CTX* bn_ctx = NULL;
	BN_MONT_CTX* bn_mon_ctx = NULL;

	orderBn = BN_new();
	if (!orderBn)
		goto error;
	
	bn_ctx = BN_CTX_new();
	bn_mon_ctx = BN_MONT_CTX_new();
	if (!bn_ctx || !bn_mon_ctx)
		goto error;

	const EC_GROUP* group = EC_KEY_get0_group(key);
	if (!group)
		goto error;

	if (!EC_GROUP_get_order(group, orderBn, NULL))
		goto error;
	EC_POINT* generator = EC_POINT_new(group);;

	BN_MONT_CTX_set(bn_mon_ctx, orderBn, bn_ctx);

	m = BN_bin2bn(sha256, sha256len, NULL);
	BN_mod(m, m, orderBn, bn_ctx);

	//Generate R in order to get custom "random number"
	sig_r = BN_new();
	EC_POINT_mul(group, generator, m, NULL, NULL, bn_ctx);
	EC_POINT_get_affine_coordinates(group, generator, sig_r, NULL, bn_ctx);
	BN_nnmod(sig_r, sig_r, orderBn, bn_ctx);

	//Generate M^-1
	exp = BN_new();
	minv = BN_new();
	if (!exp || !minv)
		goto error;

	BN_set_word(exp, (BN_ULONG)2);
	BN_sub(exp, orderBn, exp);
	BN_mod_exp_mont(minv, m, exp, orderBn, bn_ctx, bn_mon_ctx);

	*kinv = minv; *kinvlen = BN_num_bytes(minv);
	*rp = sig_r; *rplen = BN_num_bytes(sig_r);
	
	rc = 1;
error:
	if (m) BN_free(m);
	if (orderBn) BN_free(orderBn);
	if (exp) BN_free(exp);
	if (bn_ctx) BN_CTX_free(bn_ctx);
	if (bn_mon_ctx) BN_MONT_CTX_free(bn_mon_ctx);
	return rc;	
}
